/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mitaka.h"
#include  "mtk_planetarium.h"
#include  "mtk_switches.h"
#include  "mtk_stars.h"
#include  "mtk_constellation.h"
#include  "mtk_solar_system.h"
#include  "mtk_sun.h"
#include  "mtk_planets.h"
#include  "mtk_satellites.h"
#include  "mtk_planet_position.h"
#include  "mtk_planet_orientation.h"
#include  "mtk_planet_phys_params.h"
#include  "mtk_atmosphere.h"
#include  "mtk_objects.h"
#include  "mtk_target.h"
#include  "mtk_camera_setting.h"
#include  "mtk_draw_struct.h"
#include  "mtk_font.h"
#include  "mtk_text.h"
#include  "mtk_gl_common.h"
#include  "../common/graphics/3D_common.h"
#include  "../common/graphics/gl_common.h"
#include  "../common/bilingual.h"


//-----------
//  Externs
//-----------
extern GLuint  dlClstSphere;
extern float  landing_lon, landing_lat;
extern TopoSphere	tpEarth, tpMars;


void	draw_celestial_grid(const mtk_draw_struct& ds, bool bDigit);
void	draw_ecliptic(const mtk_draw_struct& ds);
void	draw_galactic_grid(const mtk_draw_struct& ds, bool bDigit);
void	draw_milkyway(const mtk_draw_struct& ds);


//-----------
//  Defines
//-----------

//--  NGC data  --
struct NGC_data2 : public NGC_data
{
	char	name[40];
};



//-------------
//  Constants
//-------------
bilingual bszCompass[8] = {
	{ "S", "{  |݂Ȃ}" }, { "SE", "{쓌|ȂƂ}" }, { "E", "{|Ђ}" }, { "NE", "{k|قƂ}" },
	{ "N", "{k|}" },     { "NW", "{k|ق}" }, { "W", "{|ɂ}" },   { "SW", "{쐼|Ȃ񂹂}" }
};

const RGBAf		col_NGC(0.2,0.9,0.5,0.8);			// NGC



//-----------
//  Objects
//-----------
map<short, NGC_data2>		NGCdata, ICdata;	// NGC

static float  g_plnt_angle = plnt_angle_ini;	// p	
static float  g_plnt_R;						// ka
static float  g_plnt_d = 0;					// k̒Ŝ


// _̕ʂƍx
static float g_plnt_long = 180;
static float g_plnt_lati = plnt_lati_ini;

// C̖邳
float AtmosphereIntensityPlanetarium;
float AtmosphereIntensityPlanetariumDef = 20.0;


float	get_plnt_R0()		{ return 0.5 * get_z_far(); }

float	get_plnt_angle()	{ return g_plnt_angle; }
float	get_plnt_R()		{ return g_plnt_R; }
float	get_plnt_d()		{ return g_plnt_d; }
void	set_plnt_R(float R)	{ g_plnt_R = R; }
float	get_plnt_long()		{ return g_plnt_long; }
float	get_plnt_lati()		{ return g_plnt_lati; }
void	set_plnt_long(const double& lon)	{ g_plnt_long = lon;  }
void	set_plnt_lati(const double& lat)	{ g_plnt_lati = lat; }

float	get_planetarium_zooming_factor()	{ return  get_plnt_R() / get_plnt_R0(); }



//------------
//  NGC data
//------------
int
read_a_NGC_record(FILE* fp, NGC_data2& nd)
{
	//--- read data
	int count;
	count = fread(&nd.NGC_NO,	sizeof(nd.NGC_NO),	1, fp);
	if (count < 1)  return FILE_FAILED;
	fread(&nd.bNGC,	sizeof(nd.bNGC),	1, fp);
	fread(&nd.eqPos.RAdeg,	sizeof(nd.eqPos.RAdeg),		1, fp);
	fread(&nd.eqPos.DEdeg,	sizeof(nd.eqPos.DEdeg),		1, fp);
	fread(&nd.type,	sizeof(nd.type),	1, fp);
	fread(&nd.size,	sizeof(nd.size),	1, fp);
	fread(&nd.mag,	sizeof(nd.mag),		1, fp);
	fread(&nd.bPhotographicMag,	sizeof(nd.bPhotographicMag),	1, fp);
	nd.name[0] = '\0';
	return  FILE_SUCCESS;
}

void
set_NGC_name(FILE* fp)
{
	const int record_size = 70+1;
	char  buf[record_size+1+1];
	int count;

	while (true) {
		//--- read one line
		buf[0] = 0;
		if (fgets(buf+1, record_size, fp) == NULL)  return;

		//--- read data
		short no;
		count = sscanf(buf+38, "%4hd", &no);
		if (count == 0)  continue;

		map<short, NGC_data2>::iterator it;
		if (buf[37]!='I') {
			it = NGCdata.find(no);
			if (it==NGCdata.end())  continue;
		} else {
			it = ICdata.find(no);
			if (it==ICdata.end())  continue;
		}

		char* p = buf+35;
		while (*p==' ') { p--; }
		p++;
		*p = '\0';
		strcpy(it->second.name, buf+1);
	}
}


bool
read_NGC_data(FILE* fp)
{
	int  res;
	NGC_data2  nd;

	while (true) {
		res = read_a_NGC_record(fp, nd);
		if (res == FILE_FAILED)  break;
		if (res == FILE_SUCCESS)	{
			if (nd.bNGC)	NGCdata.insert( pair<short,NGC_data2>(nd.NGC_NO, nd) );
			else			ICdata.insert(  pair<short,NGC_data2>(nd.NGC_NO, nd) );
		}
	}

	return  true;
}




//-------------------------------
//  vl^E̎pݒ肷
//-------------------------------
void
set_planetarium_angle(float angle)
{
	g_plnt_angle = angle;
	const float th0 = plnt_angle_ini/2 / 180 * pi;
	const float th  = g_plnt_angle /2 / 180 * pi;
	const float R0 = get_plnt_R0();
	float R, d;
	R = R0 / (sin(th)/tan(th0) + 1 -cos(th));
	d = R - R0;
	g_plnt_R = R;
	g_plnt_d = d;
}



//----------------
//  k
//----------------
void
draw_grid_compass(const float r, const float sz)
{
	const float  a = pi / 180.0f;
	const float  b = pi / 12.0f;

	for (int i=0; i<8; i++) {
		float th_d,phi_d;
		th_d  = 0 + 0.3;

		phi_d = float(i)*45;
		if (i==4)		phi_d += 1.5;	// k
		if (i%2 == 1)	phi_d += 1.5;
		
		glPushMatrix();
			glRotatef(phi_d-90,0,0,1);
			glRotatef(th_d,1,0,0);
			glTranslatef(0,r,0);
			glRotatef(90, 1,0,0);

			glScalef(sz,sz,sz);
			
			draw_bilingual(bszCompass[i]);
		glPopMatrix();
	}


	// ܓx
	for (int i=-90; i<=90; i+=15) {
		float th_d,phi_d;
		th_d  = float(i) + 0.3;
		phi_d = 0 -0.3;
		
		glPushMatrix();
			glRotatef(phi_d-90 +180,0,0,1);
			glRotatef(th_d,1,0,0);
			glTranslatef(0,r,0);
			glRotatef(90, 1,0,0);

			glScalef(sz,sz,sz);
			draw_textf("%d", i);
		glPopMatrix();
	}
}



//--------------------
//  nW
//--------------------
void
go_to_horizon_coord(const mtk_draw_struct& ds)
{
	set_planet_orientation( get_object(ds.tgt_ID).get_index() );
	//set_planet_orientation(ds.tgt.id2);

	glRotatef(landing_lon, 0,0,1);
	glRotatef(90-landing_lat, 0,1,0);
}

void
draw_horizon_coord(const mtk_draw_struct& ds)
{
	int tgt_ty = get_object( ds.tgt_ID ).get_type();
	if (tgt_ty != OBJ_TYPE_PLANET && tgt_ty != OBJ_TYPE_SATELLITE)	return;

	const float r = 1.00001;


	const float lw = 1.0 / (ds.rad_per_pixel / 0.001);
	glLineWidth(lw);

	glColor4f(0,0.7,0.9,0.8f);
	glPushMatrix();
		go_to_horizon_coord(ds);
		draw_grid_compass(r, 1.0f/40.0f);
		glScalef(r,r,r);
		glCallList(dlClstSphere);
	glPopMatrix();
}



//----------
//  `
//----------
const float sky_doom_fac = 0.95;

void
draw_sky2(const mtk_draw_struct& ds, const vector3f& n_sun, bool bMultipleScattering)
{
	const int imax = 30;
	const int jmax = 30;
	RGBAf rgba[imax+1][jmax+1];
	vector3f  pos[imax+1][jmax+1];

	// parameter for atmosphere
	const float zrt = get_zoom_rate().zmpPlanet;
	const float R1 = get_planet_Re(ds.tyMode, 2) * 1000;
	const float H  = 8000 * zrt;
	const float R2 = R1 + 20*H;
	const float n0 = 2.69e25 / zrt;	// Earth
	const float I0 = AtmosphereIntensityPlanetarium;	// incident intensity


	//---  ϑn_  ---
	double elv = 0.0;
	if ( get_object( ds.tgt_ID ).get_index() == 2 && tpEarth.is_loaded()) {
		elv = tpEarth.get_elevation(landing_lon, landing_lat);
	}
	const vector3f r0 = vector3f(0,0,R1+elv);


	// zɏƂ炳ꂽCȂꍇ͉Ȃ
	const float thc = 2 * acos(R1/R2);
	const float th  = asin(-n_sun.z);

	if (th > thc)  return;

	//
	GLfloat M[16];
	glGetFloatv( GL_MODELVIEW_MATRIX,  M);
	const vector3f ex  = normalize( vector3f(M[0], M[4], M[8]) );
	const vector3f ey  = normalize( vector3f(M[1], M[5], M[9]) );
	const vector3f ez  = normalize( vector3f(M[2], M[6], M[10]));
	const vector3f _ex = normalize( vector3f(M[0], M[1], M[2]) );
	const vector3f _ey = normalize( vector3f(M[4], M[5], M[6]) );
	const vector3f _ez = normalize( vector3f(M[8], M[9], M[10]));

	// p
	const frustum& frs = ds.frs;
	const float z1 = frs.get_z1();
	const float x1 = frs.get_x1()/z1;
	const float x2 = frs.get_x2()/z1;
	const float y1 = frs.get_y1()/z1;
	const float y2 = frs.get_y2()/z1;
	const float R  = ds.plnt_R;
	const float R0 = get_plnt_R0();
	const float th0 = 0.5 * plnt_angle_ini / 180.0 * M_PI;
	const float th1 = 0.5 * ds.plnt_angle  / 180.0 * M_PI;
	const float h = R0*cos(th0);
	const float d = (R - R0)/h;

	//-- Make color table
	const float xl = fabs(x2-x1);
	const float yl = fabs(y2-y1);
	const float l1 = min(xl, yl);
	const float l2 = max(xl, yl);
	const float tanth = tan( 0.5*ds.plnt_angle /180 * M_PI );
	const float angle_max = 2 * atan(l2/l1*tanth) / M_PI * 180;
	const float opq_fac = 14;	// 邳sx߂q

	if (angle_max < 90) {
		// 삪90xȉ

		for (int j=0; j<=jmax; j++) {
			float f = float(j)/jmax;
			float x = (x2-x1)*f + x1;
			float _y1 = y1;
			
			for (int i=0; i<=imax; i++) {
				f = float(i)/imax;
				float y = (y2-_y1)*f + _y1;

				// xNǧvZ
				vector3f nv;
				const float _R = sky_doom_fac * R/h;
				vector3f nv0 = normalize( vector3f(x,y,-1) );

				vector3f _r0(0,0,-d);

				// XN[̉]̕␳
				{
					vector3d  eF, eU, eR;
					quaterniond  ori = get_local_camera_orientation() * get_base_camera_ori();
					quaternion_to_axes( ori.inv(), eF, eU, eR);
					_r0 = eF * d;
				}
				
				const float b = dot_product(nv0, _r0);
				const float D = b*b - d*d + _R*_R;
				if (D >= 0.0) {
					const float t = -b + sqrt(D);
					nv = normalize( t*nv0 + _r0 );
				} else {
					nv.set(0,0,0);
				}

				vector3f ev;
				ev = ex*nv.x + ey*nv.y + ez*nv.z;
				if (ev.z <= 0.00001) {
					ev.z = 0.00001;
				}
				ev = normalize(ev);
				
				pos[i][j] = ev;
				const vector3f lv = n_sun;
				
				RGBf col = calc_scattered_light_intensity(R1, R2, n0, H, I0, r0, ev, lv, 22);
				if (bMultipleScattering) {
					//--  dÚAXP[̒nɂČvZĂ̂ r0/zrt Ƃ
					col += I0 * calc_multiple_scattered_light_intensity(r0/zrt, ev, lv);
				}
				
				rgba[i][j].rgb = correct_RGB(col);
				const RGBf& rgb = rgba[i][j].rgb;
				float aa = opq_fac * max(rgb.r, max(rgb.g, rgb.b));
				if (aa > 1) aa = 1;
				rgba[i][j].a   = aa;
			}
		}
	} else {
		// 삪90xȏ
		for (int i=0; i<=imax; i++) {
			float f = float(i)/imax;
			const float th = 0.5*f*M_PI;
			const float z = cos(th);
			const float rho = sin(th);
			for (int j=0; j<=jmax; j++) {
				f = float(j)/jmax;
				const float phi = 2 * f * M_PI;
				const float x = rho * cos(phi);
				const float y = rho * sin(phi);

				vector3f ev(x,y,z);
				if (ev.z <= 0.0001) {
					ev.z = 0.0001;
				}
				ev = normalize(ev);

				pos[i][j] = ev;
				const vector3f lv = n_sun;
				
				RGBf col = calc_scattered_light_intensity(R1, R2, n0, H, I0, r0, ev, lv, 22);
				if (bMultipleScattering) {
					//--  dÚAXP[̒nɂČvZĂ̂ r0/zrt Ƃ
					col += I0 * calc_multiple_scattered_light_intensity(r0/zrt, ev, lv);
				}

				rgba[i][j].rgb = correct_RGB(col);

				const RGBf& rgb = rgba[i][j].rgb;
				float aa = opq_fac * max(rgb.r, max(rgb.g, rgb.b));
				if (aa > 1) aa = 1;
				rgba[i][j].a   = aa;
			}
		}
	}


	//--
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);

	for (int i=0; i<imax; i++) {
		glBegin(GL_TRIANGLE_STRIP);
		for (int j=0; j<=jmax; j++) {
			glColor4fv(rgba[i+1][j].get_pointer());
			glVertex3fv(pos[i+1][j].get_pointer());
			glColor4fv(rgba[i  ][j].get_pointer());
			glVertex3fv(pos[i  ][j].get_pointer());
		}
		glEnd();
	}

	glDisable(GL_BLEND);
}



//------------
//  nʂ
//------------
void
draw_ground()
{
	vector3f  pos[19][73];
	glColor3f(0,0,0);

	//-- Make pos table
	for (int i=0; i<=18; i++) {
		float  th= 5*float(i) * M_PI / 180.0;
		float z = -cos(th);
		float rho = sin(th);
		for (int j=0; j<=72; j++) {
			float phi = 5*float(j) * M_PI/180.0;
			pos[i][j].set(rho*cos(phi), rho*sin(phi), z);
		}
	}

	//-- nʂ
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glBegin(GL_QUADS);
	for (int i=0; i<18; i++) {
		for (int j=0; j<72; j++) {
			glVertex3fv(pos[i+1][j].get_pointer());
			glVertex3fv(pos[i+1][j+1].get_pointer());
			glVertex3fv(pos[i][j+1].get_pointer());
			glVertex3fv(pos[i][j].get_pointer());
		}
	}
	glEnd();

	// k̃x
	glColor3f(0.6,0.6,0.5);
	for (int i=0; i<8; i++) {
		float th_d,phi_d;
		th_d  = 0 - 3;
		phi_d = float(i)*45;
		
		glPushMatrix();
			glScalef(0.99,0.99,0.99);
			glRotatef(phi_d-90,0,0,1);
			glRotatef(th_d,1,0,0);
			glTranslatef(0,1,0);
			glRotatef(90, 1,0,0);

			glScalef(0.04,0.04,0.04);
			draw_bilingual_centering(bszCompass[i]);
		glPopMatrix();
	}
}


void
draw_sky(const mtk_draw_struct& ds)
{
	const float d2r = M_PI/180.0;

	const mtk_object&  tgt = get_object(ds.tgt_ID);

	if (tgt.get_type() != OBJ_TYPE_PLANET) return;
	const float r = sky_doom_fac;
	const int plnt_idx = tgt.get_index();

	//--  nWnł̑z̈ʒuvZ  --
	vector3d rs, n_sun;
	rs = -get_planet_position(plnt_idx);						// fSWł̑z̈ʒu
	rs = get_planet_ec_to_pl_matrix(plnt_idx) * rs;				// ec -> pl 
	rs = Ry3deg(-90+landing_lat) * Rz3deg(-landing_lon) * rs;	// pl -> nW
	n_sun = rs.normalized();

	glPushMatrix();
		go_to_horizon_coord(ds);
		glScalef(r,r,r);
		if (plnt_idx == 2 && (AtmosphereMode != ATMOSPHERE_NONE)) {	// n̂݋`悷
			draw_sky2(ds, n_sun, (AtmosphereMode == ATMOSPHERE_MULTIPLE));
		}
		glScalef(0.95,0.95,0.95);
		draw_ground();
	glPopMatrix();
}



//-------------
//  NGC ̕\
//-------------
void
draw_NGC(const mtk_draw_struct& ds)
{
	char str[100];
	const float R = 0.9955;
	const float  a = pi / 180.0f;
	const float screen_w = get_screen_w();
	const float screen_z = get_screen_z();


	glColor4fv(col_NGC.get_pointer());
	glPointSize(3.5f);

	// ̒S̍W𓾂
	float M[16];
	vector3f  rc;
	glGetFloatv( GL_MODELVIEW_MATRIX, M);
	rc.x = M[12];
	rc.y = M[13];
	rc.z = M[14];

	for (map<short, NGC_data2>::const_iterator it=NGCdata.begin(); it!=NGCdata.end(); ++it) {
		const NGC_data2& nd = it->second;

		if (nd.name[0] == '\0')  continue;
		
		vector3f  xyz = nd.eqPos.get_dirc_vec();
		vector3f  xyzE = (M_R2E * xyz) * R;
		glBegin(GL_POINTS);
			glVertex3fv( xyzE.get_pointer() );			
		glEnd();


		glPushMatrix();
			glTranslatef( xyzE.x, xyzE.y, xyzE.z );
			prepare_3Dtext_2(rc, 0, 3*ds.ltf, screen_w, screen_z, true, false);
			
			if (nd.name[0]!='\0')	{ sprintf(str, " NGC%hd(%s)", nd.NGC_NO, nd.name);
			} else					{ sprintf(str, " NGC%hd", nd.NGC_NO); }

			if (nd.name[0] != '\0') {
				draw_text(str);
			}
		glPopMatrix();
	}
}



//----------------
//  ̖O̕\
//----------------
void
draw_star_name_planetarium(const mtk_draw_struct& ds)
{
	if (!(swStar & (STAR_NAME | STAR_BAYER | STAR_FRAMSTEED | STAR_HIP)))  return;

	const double R = 0.995;
	const vector3d camera_pos = ds.camera.get_pos();
	const float screen_w = get_screen_w();
	const float screen_z = get_screen_z();


	// ̒S̍W𓾂
	float M[16];
	vector3f  rc;
	glGetFloatv( GL_MODELVIEW_MATRIX, M);
	rc.x = M[12];
	rc.y = M[13];
	rc.z = M[14];

	glDisable(GL_DEPTH_TEST);

	const int N = get_stars_num();
	for (int i=0; i<N; i++) {
		const mtk_star& st = get_star(i);
		const char* name_L = st.get_name_L();
		const char* name_S = st.get_name_S();
		const bool bNameLarge = (name_L[0] != '\0');
		const bool bNameSmall = (name_S[0] != '\0');
		if (!bNameLarge && !bNameSmall)		continue;	// \閼Oꍇ

		vector3d xyz = R * normalize(st.get_pos() - camera_pos);
	
		//**** culling

		glPushMatrix();
			glTranslatef(xyz.x, xyz.y, xyz.z);

			const float sz = 0.2 * st.get_disp_size();	//******
			prepare_3Dtext_2(rc, sz, 3.3*ds.ltf, screen_w, screen_z, true, false);
    			
			glColor4fv(col_star_name.get_pointer());
			if (bNameLarge) {
				draw_text(name_L);
			}

			if (bNameSmall) {
				glColor3f(0.7,0.7,0.6);
				if (bNameLarge)		draw_textf("(%s)", name_S);
				else				draw_text(name_S);
			}
		glPopMatrix();
	}

	glEnable(GL_DEPTH_TEST);
}



void
draw_planetarium(mtk_draw_struct ds)
{
	ds.set_far_region();
	gl_set_frustum(ds.frs);
	glMatrixMode(GL_MODELVIEW);
	

	//---  Y[CEY[AEg  ---
	vector3d eF, eU, eR;
	ds.camera_base.get_axes(eF, eU, eR);
	double l = -get_plnt_d();
	glTranslated(l*eF.x, l*eF.y, l*eF.z);

	//---  ṼTCYɃXP[  ---
	const float scl = ds.plnt_R;
	glScalef(scl, scl, scl);

	prepare_draw_stars_planetarium(ds);

	draw_milkyway(ds);

	if (swConstellation & CNST_BND)		{ draw_constellation_boundary(ds); }
	if (swConstellation & CNST_ILLUST)	{ draw_constellation_illust(ds); }
	if (swConstellation & CNST_LINE)	{ draw_constellation_lines(ds); }
	if (bCelestialGrid)					{ draw_celestial_grid(ds, true); }
	if (bEcliptic)						{ draw_ecliptic(ds); }
	if (bGalacticGrid)					{ draw_galactic_grid(ds, true); }
	if (bHorizonCoord)					{ draw_horizon_coord(ds); }

	//***
	if ( test_disp_switch(DISPSW_PLANET, DISPM_DISP) )			{ draw_sun_planetarium(ds); }

	if ( ds.tgt_pos.norm() < 0.1) {
		if ( test_disp_switch(DISPSW_PLANET,	DISPM_DISP) )	{ draw_planets_planetarium(ds); }
		if ( test_disp_switch(DISPSW_SATELLITE, DISPM_DISP) )	{ draw_satellites_planetarium(ds); }
		if (bNGC)												{ draw_NGC(ds); }
		if (swConstellation & CNST_NAME)						{ draw_constellation_name(ds); }
	}

	draw_star_name_planetarium(ds);
	if (swStar & STAR_DISP)										{ draw_stars_2(ds, false, true); }

	draw_sky(ds);
}
