/**************************************************************************************************

Copyright (c) 2007   Tsunehiko Kato  and  4D2U Project, National Astronomical Observatory of Japan

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

ȉɒ߂ɏ]A{\tgEFAъ֘Ãt@Ciȉu\tgEFAvj̕擾
邷ׂĂ̐lɑ΂A\tgEFA𖳐ɈƂ𖳏ŋ܂Bɂ́A\tgEFA̕
gpAʁAύXAAfځAЕzATuCZXA/܂͔̔錠Aу\tgEFA
񋟂鑊ɓƂ錠Ɋ܂܂܂B

L̒쌠\і{\A\tgEFÂׂĂ̕܂͏dvȕɋLڂ̂Ƃ
B

\tgEFÁû܂܁vŁAł邩Öقł邩킸A̕ۏ؂Ȃ񋟂܂B
łۏ؂Ƃ́AiA̖ړIւ̓KAьNQɂĂ̕ۏ؂܂݂܂AɌ
̂ł͂܂B ҂܂͒쌠҂́A_sׁAs@sׁA܂͂ȊOł낤ƁA\t
gEFAɋN܂͊֘AA邢̓\tgEFA̎gp܂͂̑̈ɂĐ؂̐A
QȂ̋`ɂĉ̐ӔCȂ̂Ƃ܂B
***************************************************************************************************/
#include  "mtk_sun.h"
#include  "mtk_mode.h"
#include  "../common/astro/astrometry.h"
#include  "mtk_params.h"
#include  "mtk_gl_common.h"
#include  "mtk_OpenGL.h"
#include  "mtk_objects.h"
#include  "mtk_solar_system.h"
#include  "mtk_label.h"
#include  "mtk_switches.h"
#include  "mtk_planetarium.h"
#include  "mtk_dialog_box.h"
#include  "mtk_font.h"






//
const double sun_draw_scale_max = 5e5;

const bilingual bszSun			= { "Sun",			"{z|悤}" };
const bilingual bszSolarSystem	= { "Solar System",	"{zn|悤}" };



//--------------------
//  Color / Textures
//--------------------
const float  B_V_sun = 0.65;
const float  Tc_sun = B_V_to_Tc(B_V_sun);
const RGBf  col_sun( correct_RGB( blackbody_color(Tc_sun) ) );
static GLuint	texSun;

const char* szSunGlareColorDef = "#FFF2EE";
static RGBf colSunGlare;




double
get_sun_radius(int mode)
{
	if (mode == MODE_3D) {
		return R_sun_km * _1km_pc * get_zoom_rate().zmhSun;
	} else {
		return R_sun_km * _1km_pc * get_zoom_rate().zmpSunMoon;
	}
}

double
get_sun_glare_radius(int mode)
{
	return  get_sun_radius(mode) * get_fparam(MTK_FLOAT_SUN_GLARE_SIZE);
}





//------------
//  Register
//------------
void
register_object_sun()
{
	mtk_object  obj;

	//  register object
	obj.init(OBJ_TYPE_SUN, 0, 0.0f, true);
	int obj_idx = register_object(obj, "Sun");
	
	// register label
	bilingual  name = bszSun;
	float R   = 1.4 * _1km_pc * R_sun_km;
	float thc = 5 * D2R;	//**

	add_label( name.text_J, name.text_E, obj_idx, MTK_COLOR_NAME_SPECIAL_OBJECT, 1.0f, get_special_object_name_size(), R, thc, true, "SunLabel");
}



//-----------------
//  Load Textures
//-----------------
void
load_sun_textures()
{
	SetSplashText("zƘf̃eNX`Ǎ...", "Loading textures of Sun and planets...");
	texSun = LoadTexture("sun.jpg", TexShrinkRate);
}



//------------------
//  Landing radius
//------------------
double
get_landing_radius_sun()
{
	return  get_sun_radius( MODE_3D );
}



//----------------
//  Prepare draw
//----------------
bool
prepare_draw_sun(const mtk_draw_struct& ds, double* pRb, double* pRh)
{
	if (ds.scale >= sun_draw_scale_max)  return false;

	*pRb = get_sun_glare_radius(ds.tyMode);
	*pRh = get_sun_radius(ds.tyMode);

	return  true;
}



//------------
//  Position
//------------
vector3d
get_sun_position()
{
	return  vector3d(0,0,0);
}



//---------
//  Scale
//---------
double
get_scale_sun()
{
	return  30.0 * get_sun_radius(MODE_3D);
}



//---------
//  Label
//---------
bool
is_sun_label_visible()
{
	return  test_disp_switch(DISPSW_PLANET, DISPM_NAME);
}



//---------------
//  Label alpha
//---------------
float
get_sun_label_alpha(const mtk_draw_struct& ds)
{
	const float cnorm = ds.cnorm;
	float af;

	float scl_max = 5e4;
	if (cnorm >= scl_max)  return 0;

	if (cnorm >= 3e2)  {
		af = 1 - cnorm/scl_max;
		if (af <= 0) return 0;
		af *= af;
	} else {
		af = 1.0;
	}

	return  af;
}



//---------
//  Glare
//---------
void
set_sun_glare_color(const RGBf& col)
{
	colSunGlare = col;
}






//--------
//  Draw
//--------
void
draw_sun0(const float& R, const RGBf& col)
{
	float mat_d[] = {1,1,1,1}, mat_a[] = {1,1,1,1};
	mat_d[0] = col.r;
	mat_d[1] = col.g;
	mat_d[2] = col.b;
	mat_a[0] = col.r;
	mat_a[1] = col.g;
	mat_a[2] = col.b;

	glMaterialfv( GL_FRONT_AND_BACK , GL_DIFFUSE,   mat_d );
	glMaterialfv( GL_FRONT_AND_BACK , GL_AMBIENT,   mat_a );

	glEnable(GL_NORMALIZE);
	glEnable(GL_LIGHTING);
	glDisable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	glColor3f(1,1,1);
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texSun);
	glDisable(GL_BLEND);

	// _̈ʒuƖƂ炷
	glPushMatrix();
		toward_eye_pos();
		float pos[4];
		pos[0] = 0;
		pos[1] = 0;
		pos[2] = 1;
		pos[3] = 0.0f;
		glLightfv( GL_LIGHT1, GL_POSITION, pos );
	glPopMatrix();

	glPushMatrix();
		glScalef(R, R, R);
		gl_draw_LOD_sphere();
	glPopMatrix();

	glDisable(GL_LIGHTING);
	glDisable(GL_LIGHT1);
	glEnable(GL_LIGHT0);
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_NORMALIZE);
}

void
draw_sun_glare(const float& Rd, const float alp)
{
	// OA̕\
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	glDepthMask( GL_FALSE );

	glPushMatrix();
		glScalef(Rd, Rd, Rd);
		toward_eye_pos2();
		
		glBegin(GL_TRIANGLE_FAN);
		glColor4f(colSunGlare.r, colSunGlare.g, colSunGlare.b, 0.7*alp);
		glVertex3f(0,0,0);
		for (int i=0; i<=360; i+=20) {
			float x = cos(D2R*i);
			float y = sin(D2R*i);
			glColor4f(colSunGlare.r, colSunGlare.g, colSunGlare.b, 0);
			glVertex3f(x,y,0);
		}
		glEnd();
	glPopMatrix();

	glDepthMask( GL_TRUE );
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}




void
draw_sun(const mtk_draw_struct& ds, int idx, const vector3d& rpos, const double& scl)
{
	const float scale			= ds.scale;

	float af;
	if (scale >= 3e4)  {
		af = 1 - scale / sun_draw_scale_max;
		if (af <= 0) return;	//*****
	} else {
		af = 1.0;
	}
	
	RGBAf rgba;
	rgba.rgb = col_sun;
	rgba.a   = af;


	// a
	const float Re = scl * get_sun_radius(ds.tyMode);
	double th = Re/rpos.norm();
	bool bPoint = (th < 2*0.001);


	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glPushMatrix();
		glTranslated(rpos.x, rpos.y, rpos.z);

		//---  Draw  ---


		//-- Glare
		const double D = norm(rpos);
		float aa = 400.0 / D;	// 100pc
		if (aa > 0.02) {
			if (aa > 1) aa=1;		
			float Rd = Re * get_fparam(MTK_FLOAT_SUN_GLARE_SIZE);
			const double thc = 1.0e-2;
			if (Rd/D < thc) {
				Rd = thc * D;
			}

			draw_sun_glare(Rd, aa);
		}

		//-- Sun
		if (!bPoint) {
			draw_sun0(Re, RGBf(1,1,1));
		} else {
			glEnable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glPointSize( 5/(ds.rad_per_pixel/0.001)  );
			glBegin(GL_POINTS);
				glColor4fv( rgba.get_pointer() );
				glVertex3f(0,0,0);
			glEnd();
		}

	glPopMatrix();
}



void
draw_sun_planetarium(const mtk_draw_struct& ds)
{
	const float scale			= ds.scale;


	vector3d  rpos = get_sun_position() - ds.camera.get_pos();
	double L = rpos.norm();
	const double R0 = 0.92;
	vector3d  pos = R0 * rpos/L;
	float RR = get_sun_radius(MODE_PLANETARIUM) / L;

	RGBf col(1,1,1);
	glPushMatrix();
		glTranslated(pos.x, pos.y, pos.z);

		draw_sun0(RR, col);
		float Rd = RR * get_fparam(MTK_FLOAT_SUN_GLARE_SIZE);
		draw_sun_glare(Rd, 1.0f);

		//---  Name
		if ( test_disp_switch( DISPSW_PLANET, DISPM_NAME ) ) {
			prepare_3Dtext_1(RR*ds.plnt_R, 45, get_special_object_name_size()*ds.ltf, 1.0f, false);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_LIGHTING);
			glEnable(GL_BLEND);

			const RGBf &col = get_color(MTK_COLOR_NAME_SPECIAL_OBJECT);
			glColor3f(col.r, col.g, col.b);

			draw_bilingual(bszSun);
		}
	glPopMatrix();
}



//--------------
//  Draw label
//--------------
void
draw_sun_label(const mtk_draw_struct& ds, const RGBf& col, float af)
{
	const float cnorm = ds.cnorm;


	//*****
	glDisable(GL_LIGHTING);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);


	//---  XP[ɂ\ς  ---
	const double scl1 = 5;
	const double scl2 = 20;

	if (cnorm < scl1) {
		// z
		glColor4f(col.r, col.g, col.b, af);
		draw_bilingual(bszSun);
	} else if (cnorm > scl2) {
		// zn
		glColor4f(col.r, col.g, col.b, af);
		draw_bilingual(bszSolarSystem);
	} else {
		// ؂ւ̏ꍇ
		const double logscl = log(cnorm);
		const double logscl1 = log(scl1);
		const double logscl2 = log(scl2);
		float t;
		float af1, af2;
		t = (logscl - logscl1) / (logscl2 - logscl1);
		af1 = (1-t) * af;
		af2 = t * af;
		// z
		glColor4f(col.r, col.g, col.b, af1);
	
		glPushMatrix();
			draw_bilingual(bszSun);
		glPopMatrix();
		// zn
		glColor4f(col.r, col.g, col.b, af2);
		glPushMatrix();
			glTranslatef(0,0,0.02);
			draw_bilingual(bszSolarSystem);
		glPopMatrix();
	}
}



float
get_sun_label_size(const mtk_draw_struct& ds)
{
	const float cnorm = ds.cnorm;
	const double scl1 = 6;


	bool bFlg = (cnorm < scl1)  ?  false  :  true;

	const char* pstr;
	if (ds.lang == LANG_ENG) {
		pstr = (bFlg)  ?  bszSolarSystem.text_E  :  bszSun.text_E;
	} else {
		pstr = (bFlg)  ?  bszSolarSystem.text_J  :  bszSun.text_J;
	}

	float w, h;
	get_text_size(pstr, w, h);


	return  w;
}






void
set_light_from_sun(const vector3d& pos)
{
	vector3d n = normalize( pos );
	float lpos[4];
	lpos[0] = -n.x;
	lpos[1] = -n.y;
	lpos[2] = -n.z;
	lpos[3] = 0.0f;
	glLightfv( GL_LIGHT0, GL_POSITION, lpos );
}





//--------
//  Info
//--------
void
make_object_info_sun(const mtk_draw_struct& ds, std::vector<std::string>& text)
{
	if (ds.lang == LANG_ENG) {
		text.push_back( "Sun" );
		text.push_back( "Type: Star" );
	} else {
		text.push_back( "z" );
		text.push_back( "ށFP" );
	}
}